@@ -30,7 +30,7 @@ extended console application environment
 
 The abstract base class :class:`MainAppBase` inherits directly from the ae namespace
 class :class:`ae console application environment class <ae.console.ConsoleApp>`.
-The so inherited helper methods are useful for to configure and
+The so inherited helper methods are useful for to log, configure and
 control the run-time of your GUI app via command line arguments,
 :ref:`config-options` and :ref:`config-files`.
 
@@ -43,89 +43,6 @@ the concepts of :ref:`application status` and :ref:`application context`,
 explained further down.
 
 
-integrate new gui framework
----------------------------
-
-For to integrate a new Python GUI framework you have to declare a
-new class that inherits from :class:`MainAppBase` and implements at
-least the abstract method :meth:`~MainAppBase.init_app`.
-
-Most GUI frameworks are providing an application class that need to
-be initialized. For to keep a reference to the framework app class
-instance within your main app class you can use the
-:attr:`~MainAppBase.framework_app` attribute of :class:`MainAppBase`.
-
-If you want ot use the :attr:`~MainAppBase.root_win` and/or
-:attr:`~MainAppBase.root_layout` attributes you can initialize them
-also within :meth:`~MainAppBase.init_app`.
-
-.. hint::
-    The initialization of the attributes :attr:`~MainAppBase.framework_app`
-    :attr:`~MainAppBase.root_win` and :attr:`~MainAppBase.root_layout` is optional.
-    You could e.g. also add the framework application as mixin
-    to the main app class.
-
-A typical implementation of a framework-specific main app class
-looks like::
-
-    from new_gui_framework import NewFrameworkApp
-
-    class NewFrameworkMainApp(MainAppBase):
-        def init_app(self):
-            self.framework_app = NewFrameworkApp()
-            self.root_win = MainView()
-            self.root_layout = self.root_win.main_layout
-
-            return self.framework_app.start, self.framework_app.stop
-
-:meth:`~MainAppBase.init_app` will be executed only once
-at app startup for to initialize the GUI framework and
-prepare it for the app startup.
-
-For to finally startup the app the :meth:`~MainAppClass.run_app` method
-has to be called from the main module of your app project.
-:meth:`~MainAppBase.run_app` will then start the GUI event loop by
-calling the first method that got returned by :meth:`~MainAppBase.init_app`.
-
-
-key press event dispatching
-___________________________
-
-For to provide a on_key_press event to the applications that will use the new
-GUI framework you have to catch the key press events of the framework,
-convert/normalize them and then call the :meth:`~MainAppBase.key_press_from_framework`
-with the normalized modifiers and key args.
-
-The :paramref:`~MainAppBase.key_press_from_framework.modifiers` arg is a string
-that can contain several of the following sub-strings, always in the alphabetic
-order (like listed below):
-
-* Alt
-* Ctrl
-* Meta
-* Shift
-
-The :paramref:`~MainAppBase.key_press_from_framework.key` arg is a string
-that is specifying the last pressed key. If the key is not representing
-not a single character but a command key, then `key` will be one of the
-following strings:
-
-* escape
-* tab
-* backspace
-* enter
-* del
-* enter
-* up
-* down
-* right
-* left
-* home
-* end
-* pgup
-* pgdown
-
-
 application status
 ------------------
 
@@ -138,9 +55,10 @@ This namespace portion introduces the section `aeAppState` in the app
 for to be recovered on the next startup of your application.
 
 .. hint::
-    The section name is declared by the :data:`APP_STATE_SECTION_NAME`
-    constant. If you need to access this config section directly then
-    please use this constant instead of the hardcoded section name.
+    The section name `aeAppState` is declared by the
+    :data:`APP_STATE_SECTION_NAME` constant. If you need
+    to access this config section directly then please
+    use this constant instead of the hardcoded section name.
 
 .. _app-state-variables:
 
@@ -156,20 +74,21 @@ This module is providing/pre-defining the following application state variables:
 Which app state variables are finally used by your app project is (fully data-driven)
 depending on the app state :ref:`config-variables` detected in all the
 :ref:`config-files` that are found/available at run-time of your app.
-The names of the available application state variables can be
+The names of all the available application state variables can be
 determined with the main app helper method :meth:`~MainAppBase.app_state_keys`.
 
 If your application is e.g. supporting a user-defined font size, using the
-provided/pre-defined app state variable :attr:`~MainAppBase.font_size`, then it has to call
-the method :meth:`change_app_state` with the :paramref:`~MainAppBase.change_app_state.state_name`
-set to `font_size` every time when the user has changed the font size of your app.
+provided/pre-defined app state variable :attr:`~MainAppBase.font_size`, then
+it has to call the method :meth:`change_app_state` with the
+:paramref:`~MainAppBase.change_app_state.state_name` set to `font_size`
+every time when the user has changed the font size of your app.
 
 .. hint::
     The two built-in app state variables are :attr:`~MainAppBase.context_id` and
     :attr:`~MainAppBase.context_path` will be explained detailed in the next section.
 
-The :meth:`~MainBaseApp.load_app_states` method is called on instantiation from
-the implemented main app class for to load the values of all
+The :meth:`~MainBaseApp.load_app_states` method is called on instantiation
+from the implemented main app class for to load the values of all
 app state variables from the :ref:`config-files`, and is then calling
 :meth:~MainAppBase.setup_app_states` for pass them into their corresponding
 instance attributes.
@@ -177,7 +96,7 @@ instance attributes.
 Use the main app instance attribute for to read/get the actual value of
 a single app state variable. The actual values of
 all app state variables as a dict is determining the method
-:meth:`~MainBaseApp.retrieve_app_states` for you.
+:meth:`~MainBaseApp.retrieve_app_states`.
 
 Changed app state value that need to be propagated also to the framework
 app instance should never be set via the instance attribute, instead call
@@ -185,16 +104,19 @@ the method :meth:`~MainBaseApp.change_app_state` (which ensures (1) the
 propagation to any duplicated (observable/bound) framework property and
 (2) the event notification to related the main app instance method).
 
-For to to save the app state to the :ref:`config-files` the implementing main app
-instance has to call the method :meth:`~MainBaseApp.save_app_states` - this could be
-done e.g. after the app state has changed or at least on quiting the application.
+For to to save the app state to the :ref:`config-files` the
+implementing main app instance has to call the method
+:meth:`~MainBaseApp.save_app_states` - this could be
+done e.g. after the app state has changed or at least
+on quiting the application.
 
 .. _app-state-constants:
 
-This module is also providing some pre-defined constants that can be optionally
-used in your application in relation to the app states data store and
-for the app state config variables :attr:`~MainAppBase.font_size`
-and :attr:`~MainAppBase.light_theme`:
+This module is also providing some pre-defined constants
+that can be optionally used in your application in relation
+to the app states data store and for the app state config
+variables :attr:`~MainAppBase.font_size` and
+:attr:`~MainAppBase.light_theme`:
 
     * :data:`APP_STATE_SECTION_NAME`
     * :data:`APP_STATE_VERSION_VAR_NAME`
@@ -237,20 +159,145 @@ For to enter into a deeper/nested context you simply call the
 and `context_path` app states.
 
 
-application state events
-------------------------
+application events
+------------------
 
 The helper method :meth:`~MainAppBase.call_event` can be used to support
 optionally implemented event callback routines.
 
-This method is used internally by this module for to fire notification events when one of
-the app state variables gets changed - like e.g. the :ref:`application context` or if
+An implementation for a GUI framework should provide at least the following
+events:
+
+* on_app_start
+* on_app_stop
+
+
+application state change events
+_______________________________
+
+The method :meth:`~MainAppBase.call_event` is also used internally by this
+module for to fire notification events when one of the app state variables
+gets changed - like e.g. the :ref:`application context` or if
 the user changes the font size.
 
 The method name of the notification event consists of the prefix ``on_`` followed
 by the variable name of the app state. So on a change of the `font_size` the
 notification event ``on_font_size`` will be called (if exists as a method
 of the main app instance).
+
+
+key press events
+________________
+
+For to provide key press events to the applications that will use the new
+GUI framework you have to catch the key press events of the framework,
+convert/normalize them and then call the :meth:`~MainAppBase.key_press_from_framework`
+with the normalized modifiers and key args.
+
+The :paramref:`~MainAppBase.key_press_from_framework.modifiers` arg is a string
+that can contain several of the following sub-strings, always in the alphabetic
+order (like listed below):
+
+* Alt
+* Ctrl
+* Meta
+* Shift
+
+The :paramref:`~MainAppBase.key_press_from_framework.key` arg is a string
+that is specifying the last pressed key. If the key is not representing
+not a single character but a command key, then `key` will be one of the
+following strings:
+
+* escape
+* tab
+* backspace
+* enter
+* del
+* enter
+* up
+* down
+* right
+* left
+* home
+* end
+* pgup
+* pgdown
+
+On call of :meth:`~MainAppBase.key_press_from_framework` this method
+will try to dispatch the key press event to your application. First it will
+check the app instance if it has declared a method with the name
+`on_key_press_of_<modifiers>_<key>` and if so it will call this method.
+
+If this method does return False (or any other value resulting in False)
+then :meth:`~MainAppBase.key_press_from_framework` will check for
+a method with the same name in lower-case and if exits it will call
+this method.
+
+If also the second method does return False, then it will try to call the
+event method `on_key_press` of the app instance (if exists) with the modifiers
+and the key as arguments.
+
+If the `on_key_press` method does also return False then
+:meth:`~MainAppBase.key_press_from_framework` will finally pass the key
+press event to the original key press handler of the GUI framework
+for further processing.
+
+
+integrate new gui framework
+---------------------------
+
+For to integrate a new Python GUI framework you have to declare a
+new class that inherits from :class:`MainAppBase` and implements at
+least the abstract method :meth:`~MainAppBase.init_app`.
+
+A minimal implementation of this method would look like::
+
+    def init_app(self):
+        return None, None
+
+Most GUI frameworks are providing classes that need to
+be instantiated for the application and for each window.
+For to keep a reference to these instances within your main app class
+you can use the attributes :attr:`~MainAppBase.framework_app`
+and :attr:`~MainAppBase.framework_win` of :class:`MainAppBase`.
+
+The initialization of the attributes :attr:`~MainAppBase.framework_app`
+and :attr:`~MainAppBase.framework_win` is optional and can be
+done e.g. within :meth:`~MainAppBase.init_app`.
+
+.. note::
+    If :attr:`~MainAppBase.framework_win` is set to a window instance,
+    then the window instance has to provide
+    a `close` method, which will be called automatically by the
+    :meth:`~MainAppBase.stop_app`.
+
+A typical implementation of a framework-specific main app class
+looks like::
+
+    from new_gui_framework import NewFrameworkApp, MainWindowClassOfNewFramework
+
+    class NewFrameworkMainApp(MainAppBase):
+        def init_app(self):
+            self.framework_app = NewFrameworkAppClass()
+            self.framework_win = MainWindowClassOfNewFramework()
+
+            return self.framework_app.start, self.framework_app.stop
+
+:meth:`~MainAppBase.init_app` will be executed only once
+at the main app class instantiation. Only the main app instance
+has to initialize the GUI framework for to prepare the app startup
+and has to return at least a callable for to start the event loop
+of the GUI framework.
+
+.. hint::
+    Although not recommended because of possible namespace conflicts,
+    one could e.g. alternatively integrate the framework application class
+    as a mixin to the main app class.
+
+For to finally startup the app the :meth:`~MainAppClass.run_app` method
+has to be called from the main module of your app project.
+:meth:`~MainAppBase.run_app` will then start the GUI event loop by
+calling the first method that got returned by :meth:`~MainAppBase.init_app`.
 """
 from abc import ABC, abstractmethod
 from configparser import NoSectionError
@@ -262,7 +309,7 @@ from ae.console import ConsoleApp                   # type: ignore
 from ae.files import FilesRegister, RegisteredFile  # type: ignore
 
 
-__version__ = '0.0.14'
+__version__ = '0.0.15'
 
 
 AppStateType = Dict[str, Any]                   #: app state config variable type
@@ -297,11 +344,11 @@ class MainAppBase(ConsoleApp, ABC):
     sound_volume: float = 0.12                              #: sound volume of current app (0.0=mute, 1.0=max)
     win_rectangle: tuple = (0, 0, 800, 600)                 #: window coordinates (x, y, width, height)
 
-    # generic run-time shortcut references provided by the main app
+    # optional generic run-time shortcut references
     framework_app: Any = None                               #: app class instance of the used GUI framework
-    root_win: Any = None                                    #: app window
-    root_layout: Any = None                                 #: app root layout
+    framework_win: Any = None                               #: window instance of the used GUI framework
 
+    # optional app resources caches
     image_files: Optional[FilesRegister] = None             #: image/icon files
     sound_files: Optional[FilesRegister] = None             #: sound/audio files
 
@@ -310,9 +357,9 @@ class MainAppBase(ConsoleApp, ABC):
 
         :param console_app_kwargs:  kwargs to be passed to the __init__ method of :class:`~ae.console_app.ConsoleApp`.
         """
-        self._exit_code: int = 0            #: init by stop_app() and passed onto OS by run_app()
-        self._start_event_loop: Callable    #: callable to start event loop of GUI framework
-        self._stop_event_loop: Callable     #: callable to start event loop of GUI framework
+        self._exit_code: int = 0                    #: init by stop_app() and passed onto OS by run_app()
+        self._start_event_loop: Optional[Callable]  #: callable to start event loop of GUI framework
+        self._stop_event_loop: Optional[Callable]   #: callable to start event loop of GUI framework
 
         self.context_path = list()  # init for Literal type recognition - will be overwritten by setup_app_states()
         super().__init__(**console_app_kwargs)
@@ -321,15 +368,13 @@ class MainAppBase(ConsoleApp, ABC):
         self.load_sounds()
 
         self._start_event_loop, self._stop_event_loop = self.init_app()
-        assert callable(self._start_event_loop) and callable(self._stop_event_loop), \
-            f"MainAppBase.__init__: start/stop event handler methods have to be callable"
-
-    # abstract methods
 
     @abstractmethod
-    def init_app(self) -> Tuple[Callable, Callable]:
+    def init_app(self, **kwargs) -> Tuple[Optional[Callable], Optional[Callable]]:
         """ initialize framework app instance and root window/layout, return GUI event loop start/stop methods.
 
+        :param kwargs:  optional arguments used by app project to inject e.g. classes that got extended
+                        by the app project for to create the framework_app/framework_win instances.
         :return:        tuple of two callable, the 1st for to start and the 2nd for to stop/exit the GUI event loop.
         """
 
@@ -495,10 +540,11 @@ class MainAppBase(ConsoleApp, ABC):
         if not self._parsed_args:
             self._parse_args()
 
-        try:
-            self._start_event_loop()
-        finally:
-            self.shutdown(self._exit_code or None)  # don't call sys.exit() for zero exit code
+        if self._start_event_loop:                      # not needed for SubApp or additional Window instances
+            try:
+                self._start_event_loop()
+            finally:
+                self.shutdown(self._exit_code or None)  # don't call sys.exit() for zero exit code
 
     def save_app_states(self) -> str:
         """ save app state in config file.
@@ -547,11 +593,11 @@ class MainAppBase(ConsoleApp, ABC):
         self.dpo(f"MainAppBase.stop_app {exit_code}")
         self._exit_code = exit_code
 
-        if self.root_win:
-            # neither self.framework_app.stop() nor self.framework_app._qapp.exit(exit_code) trigger window closed event
-            self.root_win.close()   # close window to save app state data
+        if self.framework_win:
+            self.framework_win.close()      # close window to save app state data
 
-        self._stop_event_loop()     # will exit the self._start_event_loop() method called by self.run_app()
+        if self._stop_event_loop:
+            self._stop_event_loop()         # will exit the self._start_event_loop() method called by self.run_app()
 
     def _update_observable_app_states(self, app_state: AppStateType):
         """ update all the observable app states.
